# encoding=UTF-8
from __future__ import unicode_literals

import datetime

import pytz

from .logging import logger


class Confirmation(object):
    RESPONSE_CODES = {
        '0': 'Transacción aprobada.',
        '-1': 'Rechazo de tx. en B24, No autorizada',
        '-2': 'Transacción debe reintentarse.',
        '-3': 'Error en tx.',
        '-4': 'Rechazo de tx. En B24, No autorizada',
        '-5': 'Rechazo por error de tasa.',
        '-6': 'Excede cupo máximo mensual.',
        '-7': 'Excede límite diario por transacción.',
        '-8': 'Rubro no autorizado.'
    }

    def __init__(self, commerce, request_ip, data):
        self.commerce = commerce
        self.request_ip = request_ip
        self.params = self.parse(data['TBK_PARAM'])
        logger.confirmation(self)

    def parse(self, tbk_param):
        decrypted_params, signature = self.commerce.webpay_decrypt(tbk_param)
        params = {}
        for line in decrypted_params.split('#'):
            index = line.find('=')
            params[line[:index]] = line[index + 1:]
        params['TBK_MAC'] = signature
        return params

    def is_success(self):
        return self.params['TBK_RESPUESTA'] == '0'

    @property
    def acknowledge(self):
        return self.commerce.webpay_encrypt('ACK')

    @property
    def reject(self):
        return self.commerce.webpay_encrypt('ERR')

    @property
    def message(self):
        return self.RESPONSE_CODES[self.params['TBK_RESPUESTA']]

    @property
    def paid_at(self):
        fecha_transaccion = self.params['TBK_FECHA_TRANSACCION']
        hora_transaccion = self.params['TBK_HORA_TRANSACCION']
        m = int(fecha_transaccion[:2])
        d = int(fecha_transaccion[2:])
        h = int(hora_transaccion[:2])
        i = int(hora_transaccion[2:4])
        s = int(hora_transaccion[4:])

        santiago = pytz.timezone('America/Santiago')
        today = santiago.localize(datetime.datetime.today())
        santiago_dt = santiago.localize(datetime.datetime(today.year, m, d, h, i, s))

        return santiago_dt

    @property
    def amount(self):
        return int(self.params['TBK_MONTO']) / 100

    @property
    def transaction_id(self):
        return int(self.params['TBK_ID_TRANSACCION'])

    @property
    def order_id(self):
        return self.params['TBK_ORDEN_COMPRA']



import sys
import random
import hashlib
import re

from six.moves.urllib.parse import urlparse

import requests

from .commerce import Commerce
from .logging import logger
from . import TBK_VERSION_KCC

__all__ = ['Payment', 'PaymentError']

REDIRECT_URL = "%(process_url)s?TBK_VERSION_KCC=%(tbk_version)s&TBK_TOKEN=%(token)s"
PYTHON_VERSION = "%d.%d" % (sys.version_info.major, sys.version_info.minor)
USER_AGENT = "TBK/%(TBK_VERSION_KCC)s (Python/%(PYTHON_VERSION)s)" % {
    'TBK_VERSION_KCC': TBK_VERSION_KCC,
    'PYTHON_VERSION': PYTHON_VERSION
}


class Payment(object):
    _token = None
    _params = None
    _transaction_id = None

    def __init__(self, request_ip, amount,
                 order_id, success_url, confirmation_url,
                 session_id=None, failure_url=None, commerce=None,):
        self.commerce = commerce or Commerce.create_commerce()
        self.request_ip = request_ip
        self.amount = int(amount)
        self.order_id = order_id
        self.success_url = success_url
        self.confirmation_url = confirmation_url
        self.session_id = session_id
        self.failure_url = failure_url or success_url

    @property
    def redirect_url(self):
        return REDIRECT_URL % {
            'tbk_version': TBK_VERSION_KCC,
            'process_url': self.process_url(),
            'token': self.token()
        }

    def token(self):
        if not self._token:
            self._token = self.fetch_token()
            logger.payment(self)
        return self._token

    def fetch_token(self):
        validation_url = self.validation_url()
        is_redirect = True

        while is_redirect:
            response = requests.post(
                validation_url,
                data={
                    'TBK_VERSION_KCC': TBK_VERSION_KCC,
                    'TBK_CODIGO_COMERCIO': self.commerce.id,
                    'TBK_KEY_ID': self.commerce.webpay_key_id,
                    'TBK_PARAM': self.params()
                },
                headers={
                    'User-Agent': USER_AGENT
                },
                allow_redirects=False
            )
            is_redirect = response.is_redirect
            validation_url = response.headers.get('location')

        if response.status_code != 200:
            raise PaymentError("Payment token generation failed")

        body, _ = self.commerce.webpay_decrypt(response.content)

        return self.get_token_from_body(body)

    def get_token_from_body(self, body):
        TOKEN = 'TOKEN='
        ERROR = 'ERROR='
        lines = body.strip().split('\n')
        for line in lines:
            if line.startswith(TOKEN):
                token = line[len(TOKEN):]
            elif line.startswith(ERROR):
                error = line[len(ERROR):]
        if error != '0':
            raise PaymentError("Payment token generation failed. ERROR=%s" % error)
        return token

    def process_url(self):
        if self.commerce.testing:
            return "https://certificacion.webpay.cl:6443/filtroUnificado/bp_revision.cgi"
        return "https://webpay.transbank.cl:443/filtroUnificado/bp_revision.cgi"

    def validation_url(self):
        if self.commerce.testing:
            return "https://certificacion.webpay.cl:6443/filtroUnificado/bp_validacion.cgi"
        return "https://webpay.transbank.cl:443/filtroUnificado/bp_validacion.cgi"

    def params(self):
        if not self._params:
            self.verify()
            self._params = self.commerce.webpay_encrypt(self.raw_params())
        return self._params

    def verify(self):
        if self.commerce is None:
            raise PaymentError("Commerce required")
        if self.amount is None or self.amount <= 0:
            raise PaymentError("Invalid amount %s" % self.amount)
        if self.order_id is None:
            raise PaymentError("Order ID required")
        if self.success_url is None:
            raise PaymentError("Success URL required")
        if self.confirmation_url is None:
            raise PaymentError("Confirmation URL required")
        confirmation_uri = urlparse(self.confirmation_url)
        if re.match('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', confirmation_uri.hostname) is None:
            raise PaymentError("Confirmation URL host MUST be an IP address")

    def transaction_id(self):
        if not self._transaction_id:
            self._transaction_id = random.randint(0, 10000000000 - 1)
        return self._transaction_id

    def raw_params(self, splitter="#", include_pseudomac=True):
        params = []
        params += ["TBK_ORDEN_COMPRA=%s" % self.order_id]
        params += ["TBK_CODIGO_COMERCIO=%s" % self.commerce.id]
        params += ["TBK_ID_TRANSACCION=%s" % self.transaction_id()]
        uri = urlparse(self.confirmation_url)
        params += ["TBK_URL_CGI_COMERCIO=%s" % uri.path]
        params += ["TBK_SERVIDOR_COMERCIO=%s" % uri.hostname]
        params += ["TBK_PUERTO_COMERCIO=%s" % uri.port]
        params += ["TBK_VERSION_KCC=%s" % TBK_VERSION_KCC]
        params += ["TBK_KEY_ID=%s" % self.commerce.webpay_key_id]
        params += ["PARAMVERIFCOM=1"]

        if include_pseudomac:
            h = hashlib.new('md5')
            h.update(self.raw_params('&', False))
            h.update(str(self.commerce.id))
            h.update("webpay")
            mac = str(h.hexdigest())

            params += ["TBK_MAC=%s" % mac]

        params += ["TBK_MONTO=%d" % int(self.amount * 100)]
        if self.session_id:
            params += ["TBK_ID_SESION=%s" % self.session_id]
        params += ["TBK_URL_EXITO=%s" % self.success_url]
        params += ["TBK_URL_FRACASO=%s" % self.failure_url]
        params += ["TBK_TIPO_TRANSACCION=TR_NORMAL"]
        return splitter.join(params)


class PaymentError(Exception):
    pass



import os

from .encryption import Encryption, Decryption

from Crypto.PublicKey import RSA

__all__ = ['Commerce']

KEYS_DIR = os.path.join(os.path.dirname(__file__), 'keys')
TEST_COMMERCE_KEY_PATH = os.path.join(KEYS_DIR, 'test_commerce.pem')
TEST_WEBPAY_KEY_PATH = os.path.join(KEYS_DIR, 'webpay_test.101.pem')
WEBPAY_KEY_PATH = os.path.join(KEYS_DIR, 'webpay.101.pem')


class Commerce(object):
    TEST_COMMERCE_KEY = open(TEST_COMMERCE_KEY_PATH, 'r').read()
    TEST_COMMERCE_ID = "597026007976"

    webpay_key_id = 101

    def __init__(self, id=None, key=None, testing=False):
        self.testing = testing
        self.id = self.__get_id(id)
        self.key = self.__get_key(key)

    @staticmethod
    def create_commerce():
        """
        Creates commerce from environment variables.
        """
        commerce_id = os.getenv('TBK_COMMERCE_ID')
        commerce_key = os.getenv('TBK_COMMERCE_KEY')
        commerce_testing = os.getenv('TBK_COMMERCE_TESTING') == 'True'

        if not commerce_testing:
            if commerce_id is None:
                raise ValueError("create_commerce needs TBK_COMMERCE_ID environment variable")
            if commerce_key is None:
                raise ValueError("create_commerce needs TBK_COMMERCE_KEY environment variable")

        return Commerce(
            id=commerce_id or Commerce.TEST_COMMERCE_ID,
            key=commerce_key,
            testing=commerce_testing
        )

    def __get_key(self, key):
        if not key:
            if self.testing:
                return self.TEST_COMMERCE_KEY
            raise TypeError("Commerce needs a key")
        return key

    def __get_id(self, id):
        if not id:
            if self.testing:
                return self.TEST_COMMERCE_ID
            raise TypeError("Commerce needs an id")
        return id

    def webpay_decrypt(self, encrypted):
        commerce_key = self.get_commerce_key()
        webpay_key = self.get_webpay_key()
        decryption = Decryption(commerce_key, webpay_key)
        decrypted = decryption.decrypt(encrypted)
        return decrypted

    def webpay_encrypt(self, decrypted):
        commerce_key = self.get_commerce_key()
        webpay_key = self.get_webpay_key()
        encryption = Encryption(commerce_key, webpay_key)
        return encryption.encrypt(decrypted)

    def get_webpay_key(self):
        path = TEST_WEBPAY_KEY_PATH if self.testing else WEBPAY_KEY_PATH
        with open(path, 'r') as webpay_key:
            return RSA.importKey(webpay_key.read())

    def get_commerce_key(self):
        return RSA.importKey(self.key)



import os
import datetime

import pytz
from six.moves.urllib.parse import urlparse

__all__ = ['logger', 'BaseHandler', 'NullHandler']


LOG_DATE_FORMAT = "%d%m%Y"
LOG_TIME_FORMAT = "%H%M%S"


class BaseHandler(object):

    def event_payment(self, date, time, pid, commerce_id, transaction_id, request_ip, token, webpay_server):
        raise NotImplementedError("Logging Handler must implement event_payment")

    def configuration_payment(self, commerce_id, server_ip, server_port,
                              response_path, webpay_server, webpay_server_port):
        raise NotImplementedError("Logging Handler must implement configuration_payment")

    def event_confirmation(self, date, time, pid, commerce_id, transaction_id, request_ip, order_id):
        raise NotImplementedError("Logging Handler must implement event_confirmation")

    def log_confirmation(self, params, commerce_id):
        raise NotImplementedError("Logging Handler must implement log_confirmation")


class NullHandler(BaseHandler):

    def event_payment(self, **kwargs):
        pass

    def configuration_payment(self, **kwargs):
        pass

    def event_confirmation(self, **kwargs):
        pass

    def log_confirmation(self, **kwargs):
        pass


class Logger(object):

    def __init__(self, handler):
        self.set_handler(handler)

    def set_handler(self, handler):
        self.handler = handler

    def payment(self, payment):
        santiago = pytz.timezone('America/Santiago')
        now = santiago.localize(datetime.datetime.now())
        self.handler.event_payment(
            date=now.strftime(LOG_DATE_FORMAT),
            time=now.strftime(LOG_TIME_FORMAT),
            pid=os.getpid(),
            commerce_id=payment.commerce.id,
            transaction_id=payment.transaction_id,
            request_ip=payment.request_ip,
            token=payment.token(),
            webpay_server=self.get_webpay_server(payment.commerce)
        )
        response_uri = urlparse(payment.confirmation_url)
        self.handler.configuration_payment(
            commerce_id=payment.commerce.id,
            server_ip=response_uri.hostname,
            server_port=response_uri.port,
            response_path=response_uri.path,
            webpay_server=self.get_webpay_server(payment.commerce),
            webpay_server_port=self.get_webpay_port(payment.commerce)
        )

    def confirmation(self, confirmation):
        santiago = pytz.timezone('America/Santiago')
        now = santiago.localize(datetime.datetime.now())
        self.handler.event_confirmation(
            date=now.strftime(LOG_DATE_FORMAT),
            time=now.strftime(LOG_TIME_FORMAT),
            pid=os.getpid(),
            commerce_id=confirmation.commerce.id,
            transaction_id=confirmation.transaction_id,
            request_ip=confirmation.request_ip,
            order_id=confirmation.order_id,
        )
        self.handler.log_confirmation(
            params=confirmation.params,
            commerce_id=confirmation.commerce.id
        )

    def get_webpay_server(self, commerce):
        return 'https://certificacion.webpay.cl' if commerce.testing else 'https://webpay.transbank.cl'

    def get_webpay_port(self, commerce):
        return '6443' if commerce.testing else '433'

logger = Logger(NullHandler())


def configure_logger(handler):
    global logger
    logger.set_handler(handler)



import base64

from Crypto import Random
from Crypto.Hash import SHA512
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.Signature import PKCS1_v1_5

__all__ = ['Encryption', 'Decryption', 'InvalidMessageException']


class Encryption(object):
    def __init__(self, sender_key, recipient_key):
        self.sender_key = sender_key
        self.recipient_key = recipient_key

    def encrypt(self, message):
        key = self.get_key()
        iv = self.get_iv()

        encrypted_key = self.encrypt_key(key)

        signed_message = self.sign_message(message)
        encrypted_message = self.encrypt_message(signed_message, message, key, iv)

        return base64.b64encode(iv + encrypted_key + encrypted_message)

    def sign_message(self, message):
        hash = SHA512.new(message)
        signer = PKCS1_v1_5.new(self.sender_key)
        return signer.sign(hash)

    def encrypt_message(self, signed_message, message, key, iv):
        raw = str(signed_message) + str(message)
        block_size = AES.block_size
        pad = lambda s: s + (block_size - len(s) % block_size) * chr(block_size - len(s) % block_size)
        message_to_encrypt = pad(raw)
        cipher = AES.new(key, AES.MODE_CBC, iv)
        return cipher.encrypt(message_to_encrypt)

    def encrypt_key(self, key):
        public_key = self.recipient_key.publickey()
        cipher = PKCS1_OAEP.new(public_key)
        return cipher.encrypt(key)

    def get_key(self):
        return Random.new().read(32)

    def get_iv(self):
        return Random.new().read(16)


class Decryption(object):
    def __init__(self, recipient_key, sender_key):
        self.sender_key = sender_key
        self.recipient_key = recipient_key

    def decrypt(self, message):
        raw = base64.b64decode(message)

        iv = self.get_iv(raw)
        key = self.get_key(raw)

        decrypted_message = self.get_decrypted_message(iv, key, raw)

        signature = self.get_signature(decrypted_message)
        message = self.get_message(decrypted_message)

        if self.verify(signature, message):
            return message, signature

        raise InvalidMessageException("Invalid message signature")

    def get_iv(self, raw):
        return raw[:16]

    def get_key(self, raw):
        recipient_key_bytes = int(self.recipient_key.publickey().n.bit_length() / 8)
        encrypted_key = raw[16:16 + recipient_key_bytes]
        cipher = PKCS1_OAEP.new(self.recipient_key)
        return cipher.decrypt(encrypted_key)

    def get_decrypted_message(self, iv, key, raw):
        recipient_key_bytes = int(self.recipient_key.publickey().n.bit_length() / 8)
        encrypted_message = raw[16 + recipient_key_bytes:]
        unpad = lambda s: s[:-ord(s[len(s) - 1:])]
        cipher = AES.new(key, AES.MODE_CBC, iv)
        return unpad(cipher.decrypt(encrypted_message))

    def get_signature(self, decrypted_message):
        sender_key_bytes = int(self.sender_key.publickey().n.bit_length() / 8)
        return decrypted_message[:sender_key_bytes]

    def get_message(self, decrypted_message):
        sender_key_bytes = int(self.sender_key.publickey().n.bit_length() / 8)
        return decrypted_message[sender_key_bytes:]

    def verify(self, signature, message):
        hash = SHA512.new(message)
        verifier = PKCS1_v1_5.new(self.sender_key)

        return verifier.verify(hash, signature)


class InvalidMessageException(Exception):
    pass



# encoding=UTF-8
#!/usr/bin/env python
import os
from setuptools import setup, find_packages


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ''

long_description = read_file('README.md').strip().split('split here', 1)[0]

setup(
    name='tbk',
    version=__import__('tbk').__version__,
    description=' '.join(__import__('tbk').__doc__.splitlines()).strip(),
    long_description=long_description,
    author='Pedro Buron',
    author_email='pedroburonv@gmail.com',
    maintainer='Pedro Buron',
    maintainer_email='pedroburonv@gmail.com',
    url='http://github.com/pedroburon/tbk',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.7',
        'Environment :: Console',
        'Environment :: Web Environment',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    ],
    keywords=['webpay', 'chile', 'payments', 'transbank'],
    install_requires=[
        'requests>=2.3.0',
        'six>=1.7.3',
        'PyCrypto>=2.6.1',
        'pytz',
    ],
    tests_require=[
        'mock>=1.0.1',
        'nose>=1.3.3',
    ],
    packages=find_packages(),
    test_suite='nose.collector',
    zip_safe=True,
    license='GPLv2'
)
